// view.c : functions for display an image under X-window
//  a part of PMS-grabber package
// the "GNU Public Lincense" policy applies to all programs of this package
// (c) Wolfgang Koehler, wolf@first.gmd.de, Dec. 1994
//     Kiefernring 15
//     14478 Potsdam, Germany
//     0331-863238

#include "window.h"
#include <X11/Xutil.h>
#include "view.h"
#include "ximage.h"
#include "evaluation.h"
#include "grabber.h"
#include "files.h"

#include <fcntl.h>
#include <unistd.h>
#include <sys/times.h>

view_window *grab_wi;
unsigned short *cap_buf = NULL;
XImage *ximage = NULL;
mapped_image *act_image = 0;

int biWidth, biHeight;
int tcap = 0, tdraw = 0; // delta-t between two grabs: capture & draw

unsigned ticks() {  // returns time since boot in 1/100 sec ticks
  struct tms tt; 
  return times(&tt); 
}

// handles interrupt wait of frame grabber with installed irq_driver
class irq_handle {
  int fid;
  int nfrm,nmiss;
public: 
  irq_handle() {
    fid = open("/dev/fg", 0);
    if (fid > 0) printf("using interrupt driver\n"); else
      printf("interrupt driver not installed\n");
    nfrm = nmiss = 0;
  }
  ~irq_handle() {
    if (fid > 0) close(fid);
  }
  void wait_irq() {
    if (fid <= 0) return; // do nothing  
    // else : simply waits until read is performed
    nfrm++;
    int ni; read(fid,&ni,4); // read number of irqs between two read 
    static first_call = True;
    if (first_call) { ni = 0; first_call = False; }
    if (ni > 1) { 
      nmiss += ni-1; // number of lost frames
      printf("."); fflush(stdout);
    } 
    if (nmiss >= 20) { 
      // int max_fr = 100/(tcap+tdraw); compute max frame rate ???
      printf("\nframe rate too high, missed %d of %d frames\n",nmiss,nfrm);
      nmiss = nfrm = 0;  
    } 
  }
};

extern "C" {
#include "jpeg/jpeglib.h"
}

struct frame { // store one image frame
  unsigned tstamp; // time stamp
  unsigned width, height, bsize; 
  unsigned short *data;

  frame(int w, int h) : width(w), height(h) {
    bsize = 2 * width * height; // only 16bpp mode supported at the moment
    data = new unsigned short[width*height];
  }

  void capture(void *src, unsigned tx) {
    memcpy(data,src,bsize); tstamp = tx;
  }
  void load() {
    memcpy(cap_buf,data,bsize); 
    grab_wi->redraw();
  }
  void write(FILE *f, int) { // write header & compressed data 
    // fputuns(tstamp,f);  geht nicht wg buffering
    int qu = 80; // quality 
    compress_jpeg(f, qu, width, height, data); 
 
   /* save frames as single ppm files :
      char fname[80]; sprintf(fname,"pict%d.ppm",i);
      FILE *fpic = fopen(fname,"w");
      write_ppm_file_16(data,width,height,fpic);
      fclose(fpic);
   */
  }
  void read(jpeg_decompress_struct *cinfo )  { // read header & data 
    jpeg_read_header(cinfo, TRUE);
    jpeg_start_decompress(cinfo); 
    // width = cinfo->output_width; adapting not useful for greater images
    // height = cinfo->output_height;
    JPEG_decompress(cinfo, data);
    jpeg_finish_decompress(cinfo);
  }
  ~frame() { delete data; }
};

struct xpoint {
  int x, y;
  xpoint() {} // wozu ??
  xpoint(short x, short y) : x(x),y(y) {}
  void save(FILE *f) {  // save into file
    fprintf(f," %d %d",x,y);
  }
  xpoint(FILE *f) {   // constructor reads from file
    if (fscanf(f,"%d%d",&x,&y) != 2) error("reading fpoint");
  }
};

// xplist is a list of xpoints
struct xplist : public xpoint {
  xplist *rest;
  xplist(short x, short y) : xpoint(x,y) { rest = NULL; }   
  xplist(xpoint p) : xpoint(p)  { rest = NULL; }
  // push gx in front of list, return new, problematic if called for this = 0
  xplist *push(xplist *gx) { gx->rest = this; return gx; }

  xplist* operator += (xplist* xp) { xp->rest = this; return xp; }

  // delete complete list starting from this recursively
  ~xplist() { if (rest) delete rest; }

  // draw all lines of the list recursively into w,
  void draw(window *w, int xoff = 0) {
    xplist *xpr = rest; 
    if (xpr == 0) return; // end of list
    // printf("%d %d\n",this->x,this->y);
    w->line(xoff + this->x,this->y,xoff + xpr->x,xpr->y);
    xpr->draw(w,xoff);
  }

  // recursively save complete list 
  void save(FILE *f) { xpoint::save(f); if (rest) rest->save(f); } 

  // constructor reads one point from file into this
  xplist(FILE *f) : xpoint(f) { rest = NULL; } 
};

class polygon {
public:
  xplist *xp;
  int np;
  polygon() { np = 0; xp = NULL; }
  void draw(window *w, int xoff = 0) { if (xp) xp->draw(w,xoff); }
  void push(xplist *xn) { xp = xp->push(xn); np++; }
  void clear() { if (xp) delete xp; xp = NULL; np = 0; }

  void save(FILE *f) {
    fprintf(f,"%d ",np); if (xp) xp->save(f);
    fprintf(f,"\n"); // for readability
  }

  void read(FILE *f) { 
    clear();
    int n; fscanf(f,"%d",&n);
    for (; n>0; n--) push(new xplist(f));
  }
};

// a frame with some polygons attached (and methods to manipulate them)
class frame_poly : public frame {
  polygon poly[10]; // at most 10 polys per frame
  int npoly;  // actual polynom < 10
  xpoint shift;
public:
  frame_poly(int w, int h) : frame(w,h) {
    npoly = 0; shift = xpoint(100,0);
  }  
  void draw(window *w, Bool corr = False) {
    int xoff = (corr) ? 100-shift.x : 0; 
    for (int n = 0; n <= npoly; n++) poly[n].draw(w,xoff);
  }
  
  void set_shift(xpoint p) { shift = p; }
  
  void add_point(xpoint p) {
    poly[npoly].push(new xplist(p));
    draw(grab_wi);
  }
  void end_poly() {
    if (++npoly >= 10) error("too many polygons");
  }
  xpoint *last_point() {
    return poly[npoly].xp;
  }
  void load() {
    frame::load();
    draw(grab_wi);
  } 
  void poly_save(FILE *f) {
    fprintf(f,"%d %d %d\n",npoly,shift.x,shift.y);
    for (int n = 0; n <= npoly; n++) poly[n].save(f);
  }
  void poly_clear() {
    for (int n = 0; n <= npoly; n++) poly[n].clear();
    npoly = 0;
  }
  void poly_read(FILE *f) {    
    fscanf(f,"%d\n",&npoly);
    // fscanf(f,"%d %d %d\n",&npoly,&shift.x,&shift.y);
    
    for (int n = 0; n <= npoly; n++) poly[n].read(f);
  }
};

// class to show polygons of 
class poly_window : public main_window {
public:
  poly_window(int w, int h) : main_window("skelett",w,h) {
    RealizeChildren();
  }
  void show_polys(frame_poly *fr) {
    clear();
    fr->draw(this,True); // draw with correction
  }
};


#include <sys/time.h>
static void sleep(int msecs) { // sleep for msec 1/1000 secs = 10 ticks
  if (ticks <= 0) return;
  struct timeval timeout;
  timeout.tv_sec = 0; timeout.tv_usec = 1000*msecs; // microsecs 
  select(0, NULL , NULL, NULL , &timeout);
}

typedef void (*VVPVP)(void *, void *);

// new handler type for callbacks with 2 pointer arguments (type VVPVP)
// stores one pointer (this pointer) and uses the 2. as argument
struct cb_handler {
  void *vp; // usually a this pointer
  VVPVP cb;
  cb_handler(void *vp = 0, VVPVP cb = 0) : vp(vp),cb(cb) {}
  void install(void *v, VVPVP c) { vp = v; cb = c; }
  void disinstall() { cb = NULL; }
  void call(void *ev) { if (cb) cb(vp,ev); }
};

// classes "toggle_action_button" & "selection_menu" should move to window.c !!
typedef void (*VVPI)(void *, int);

// used for selection menu
class toggle_action_button : public toggle_button {
  VVPI cb;
  void *vp; int ind; 
public:
  toggle_action_button(window &parent, char * Name, int *ref, 
		       VVPI cb, void *vp, int index,
		       int w, int h, int x, int y) :
  toggle_button(parent,Name,ref,w,h,x,y), cb(cb), vp(vp), ind(index) { }
  virtual void toggle_cb(int) { cb(vp,ind); }
};

// creates a pulldown_menu with toggle buttons and a pulldown button
// the states are exclusive : only one (index = value) can be set "on"
class selection_menu {
public:
  int nb, value; // button number, actual set index of True 
  Bool *state; // only one is True
  selection_menu(menu_bar &parent, char *Name, 
		 char *bname[], int value = 0) : value(value) {
    int wm = 0, hm;     // size of pulldown_menu
    nb = 0;
    // first parse bnames to determine needed size of pulldown menu
    while (bname[nb]) wm = wm >? strlen(bname[nb++]);

    wm = (wm+2) * 6 + 10;   //  6 = length of char (+2 for picture)
    hm = nb * 20;           // 20 = height of one button
    pulldown_window * pd = new pulldown_window(wm, hm);
    state = new Bool[nb]; 
    int yp = 0; 
    for (int i=0; i < nb; i++) { 
      state[i] = (i == value);
      new toggle_action_button(*pd,bname[i], state+i, 
			       (VVPI) select_cb ,this, i, wm, 20,0, yp);
      yp += 20;
    }
    new pulldown_button(parent, pd, Name); 
  }
  void select_cb(int ind) {
    value = ind;
    for (int i=0; i < nb; i++) state[i] = (i == value);
  }
    
  ~selection_menu() { delete[] state; }
};

// play_scrollbar with additional option to act on BPress events
// here set an array element which is indexed by mode to actual value
class frame_scrollbar : public play_scrollbar {
  int *indptr;
  int *arr;
public:
  frame_scrollbar(window &parent, int w, int x, int y, int n_ticks, 
		 stepper *stp, int minp, int maxp, int *indptr, int arr[]) : 
  play_scrollbar(parent,w,x,y,n_ticks,stp,minp,maxp),indptr(indptr),arr(arr) {
  } 
  // called from interactive slider setting 
  virtual void BPress_1_CB(XButtonEvent ev) {
    play_scrollbar::BPress_1_CB(ev);
    printf("%d %d\n",value,*indptr);
    arr[*indptr] = value;
  }
};

// invoked from view_window ButtonPress
static cb_handler view_bpress, view_motion;

// main window with play_scrollbar for saving & replaying of videos
class frame_replayer : public main_window, stepper {  
  Bool video_capture;
  Bool replay_mode, wrap_mode, set_shift_mode;
  play_scrollbar *pscr;
  scrollbar *ascr; // for setting the replay delay time (in msecs)
  int nframes;
  frame_poly **frames;
  poly_window *pwin;
  char base_name[80]; // base file name for video & poly files
  int rep[3];
  int &rep_start, &rep_end; 
  // indicees of first and last frame to display when replaying
public:
  frame_replayer(char *name, int nframes, int w = 400, int h = 150)
    : main_window(name,w,h), nframes(nframes), 
      rep_start(rep[1]), rep_end(rep[2]) {
    frames = new (frame_poly*)[nframes];
    for (int i = 0; i < nframes; i++) 
      frames[i] = new frame_poly(biWidth,biHeight);
    int ww = width-20;
    menu_bar *mb = new menu_bar(*this,ww,20,10,0);
    new switch_button(*mb,"capture","stop",&video_capture);
    new toggle_button(*mb,"poll",&polling_mode);
    new instance_button <frame_replayer>(*mb,"save",video_write,this); 
    new instance_button <frame_replayer>(*mb,"read",video_read,this);
    new toggle_button(*mb,"wrap",&wrap_mode);
    char *selector[] = { "normal","start","end",0};
    selection_menu *sel = new selection_menu(*mb,"mode",selector);
    new instance_button <frame_replayer> (*mb,"close",close,this);
    new instance_button <frame_replayer> (*mb,"quit",quit,this);
    ascr = new scrollbar(*this,NULL,ww,20,10,40,0,500,80);
    int nticks = nframes-2; while (nticks > 10) nticks /= 2;
    pscr = new frame_scrollbar(*this,ww,10,70,nticks,this,0,nframes-1,
			       &sel->value, rep);
    menu_bar *mb2 = new menu_bar(*this,ww,20,10,h-30);  
    new instance_button <frame_replayer>(*mb2,"spoly",poly_save,this); 
    new instance_button <frame_replayer>(*mb2,"rpoly",poly_read,this);
    new instance_button <frame_replayer>(*mb2,"poly",toggle_poly_win,this);
    new toggle_button(*mb2,"shift",&set_shift_mode);
    new instance_button <frame_replayer>(*mb2,"clear",poly_clear,this);
    pwin = 0;
    video_capture = False; replay_mode = False;
    polling_mode = True; wrap_mode = True;
    sprintf(base_name,"default");
    rep_start = 0; rep_end = nframes;
    remap();
  } 
  void remap() { 
    view_bpress.install(this, (VVPVP) bpress);
    view_motion.install(this, (VVPVP) motion);
    exit_flag = False; 
    main_loop(); 
  }
  virtual ~frame_replayer() {
    for (int i = 0; i < nframes;i++) delete (frames[i]); 
    delete[] frames;
    view_bpress.disinstall();
  }
  
private: 
  int ifr() { return(pscr->value % nframes); } // actual frame number 

  void rubber_line(xpoint p) {  
    xpoint *org =  frames[ifr()]->last_point();
    if (org)
      XDrawLine(display,grab_wi->Win,gc_rubber,org->x,org->y,p.x,p.y);
  }
 
  xpoint pold, mark;

  void motion(XMotionEvent *ev) { // installed as cb_handler for view_window
    xpoint p(ev->x, ev->y);
    rubber_line(p); rubber_line(pold);
    pold = p; 
  }

  void bpress(XButtonEvent *ev) { // installed as cb_handler for view_window
    // printf("%d %d %d\n",ev->x,ev->y,nfr); 
    xpoint p(ev->x, ev->y); 
    frame_poly *fri = frames[ifr()]; // actual frame
    rubber_line(pold);  
    switch (ev->button) {
    case Button1: 
      if (set_shift_mode) fri->set_shift(p);
      else fri->add_point(p); 
      if (pwin) pwin->show_polys(fri);
      break;
    case Button2:
      fri->end_poly(); break;
    case Button3:   
      fri->add_point(p); 
      fri->end_poly(); break;
    }
  }
  
  char fina[80];
  // interactive open and setting base_name 
  FILE *open_file(char *defname, int mode) {
    FILE *f = open_file_interactive(defname,fina,mode);
    if (f) {
      strcpy(base_name,fina); 
      char *bp = base_name; 
      while (*bp != '.') bp++; *bp = 0; // truncate to dot (including)
    }
    return f;
  }

  void video_write() { // callback to save frames into file
    FILE *fsave = open_file("default.video",MODE_WRITE);
    if (fsave == NULL) return;
    printf("\nsaving %s ",fina);
    // start saving from actual frame (not 0) wrapped around !!! 
    for (int i=0, ifx = ifr(); i < nframes; i++,ifx++) {  
      printf("."); fflush(stdout);
      ifx %= nframes;
      frames[ifx]->write(fsave,ifx); 
    }
    printf("\n");
    fclose(fsave);
  }

  void video_read() { // callback to read frames from file
    FILE *fload = open_file("*.video",MODE_READ);
    if (fload == NULL) return;
    struct jpeg_decompress_struct cinfo;
    struct jpeg_error_mgr jerr;
    cinfo.err = jpeg_std_error(&jerr); 
    jpeg_create_decompress(&cinfo);  
    jpeg_stdio_src(&cinfo, fload);
    printf("\nloading %s ",fina);
    for (int i=0; i<nframes; i++) {
      printf("."); fflush(stdout);
      if (feof(fload)) { nframes = i; break; } // not yet consistent
      frames[i]->read(&cinfo); 
    }  
    printf("\n");
    jpeg_destroy_decompress(&cinfo);
    fclose(fload);
    replay_mode = True; polling_mode = True; // replay the loaded file 
  }
  void show_poly_win(Bool toggle) { // if True : Map <-> Unmap
    if (pwin == 0) pwin = new poly_window(biWidth,biHeight);
    XWindowAttributes attr;
    XGetWindowAttributes(display,pwin->Win,&attr);
    if (toggle && attr.map_state == IsViewable) pwin->Unmap();  
    else pwin->RealizeChildren(); 
  } 
  void toggle_poly_win() {
    show_poly_win(True);
  }
  void poly_save() { // callback to save frame polygons into file
    sprintf(fina,"%s.poly",base_name);
    FILE *f = open_file_interactive(fina,fina,MODE_WRITE);
    if (f == NULL) return;  
    printf("writing '%s'\n",fina);
    fprintf(f,"%d\n",nframes);
    for (int i = 0; i < nframes; i++) frames[i]->poly_save(f); 
    fclose(f);
  }
  
  void poly_read() { // callback to read frame polygons from file
    sprintf(fina,"%s.poly",base_name);
    FILE *f = fopen(fina,"r"); // try the same name as video file
    if (f==0) f = open_file("*.poly",MODE_READ);
    if (f==0) return; // aborted
    printf("reading '%s'\n",fina);
    int nf; fscanf(f,"%d",&nf); nf = nf <? nframes;
    for (int i = 0; i < nf; i++) frames[i]->poly_read(f); 
    fclose(f);
    replay_mode = True; polling_mode = True; // replay the loaded file 
    show_poly_win(False);
  }
  void poly_clear() { // delete all polygons
    frame_poly *fri = frames[ifr()];
    fri->poly_clear(); 
    if (pwin) pwin->show_polys(fri);
  }    
  void close() { 
    view_bpress.disinstall();
    exit_flag = True;  // return to main_loop of PMS
    Unmap(); 
  }
  void quit() { exit(0); }

  // two virtual functions from stepper : invoked from play_scrollbar
  virtual void seek_step(int i) { // seek and perform drawing
    if (polling_mode && ! replay_mode) return; // called from polling handler !
    static int told = 0; // compute time since last call  
    int delt = (ticks() - told) >? 0;
    int delay = ascr->value - delt*10 >? 0;
    if (polling_mode) sleep(delay);    
    i %= nframes; if (i < 0) i += nframes; 
    frames[i]->load();
    if (pwin) pwin->show_polys(frames[i]);
    told = ticks();
  }
  
  virtual Bool switch_play_mode(VVP /*step*/, void* /*vp*/) { 
    replay_mode = ! replay_mode; 
    polling_mode = replay_mode;
    return replay_mode;
  }

  virtual void polling_handler() {   
    int i = ifr(); 
    unsigned tx = ticks();
    if ( ! replay_mode) {
      show_capt(); // normal online grabbing
      if (video_capture) {
	frames[i]->capture(cap_buf, tx); 
	if (! wrap_mode && i == nframes - 1) 
	  video_capture = polling_mode = False; // stop 
      }
    }    
    if (replay_mode || video_capture) { 
      int i0 = replay_mode ? rep_start : 0;
      int ix = replay_mode ? rep_end : nframes;
      if (++i >= ix) i = i0; 
      pscr->change(i);  // calls seek_step indirectly !!!
    }
  }
};

// !! set this according the available RAM in your PC !!!
#define MAX_FRAMES 20

void make_video_popup() { // button callback fn
  static frame_replayer *fr = NULL;
  if (fr == NULL) fr = new frame_replayer("videos",MAX_FRAMES);
  else fr->remap();
}

// capture one frame, transform with coltab and draw into grab_wi
void show_capt() {
  if (cap_buf == NULL) return;

  // test for animations : 
  // unsigned tx = ticks();  // returns time since boot in 1/100 sec ticks
  // static unsigned tl = 0;
  // int delt = tx - tl; tl = tx;
  // printf(" %d %.2f",act_frame,0.01*delt); fflush(stdout);

  static irq_handle fg_irq;  // not global (init time too early) !!
  fg_irq.wait_irq();
  Bool rgb555 = ! True_Color; // in PseudoColor use only cap_mode 555 !!
  fast_capture(cap_buf, biWidth, biHeight, rgb555);   
      
  tcap = comp_time(); 
  // tcap + tdraw = 
  // 260x200: 0.04, 340x255: 0.06, 500x395 : 0.11 sec
  map_and_redraw();
}    

// computes the time since last call to this function
int comp_time() {
  unsigned tx = ticks();  // returns time since boot in 1/100 sec ticks
  static unsigned tl = 0;
  int delt = tx - tl; 
  tl = tx;
  return delt;
}

// uses coltab555 to directly map a rgb555-buffer into a XImage
void map_buf555(unsigned short *buf, int n) {
  unsigned char *imb =  (unsigned char *) ximage->data;
  if (imb == NULL) return;  
  int i; 
  for (i=0; i < n; i++) *imb++ = coltab555[ *buf++ ];
}

void map_and_redraw() {
  int n = biWidth*biHeight;
  if( ! True_Color) map_buf555(cap_buf, n);   
  // full color transformation : 320x240 : 0.10 sec
  grab_wi->redraw();    // 320x240 
  tdraw = comp_time(); // wird erst beim naechsten Bild dargestellt
}

// callback function to make a new mag_window and unmap it on second press
void toggle_magw() {
  static mag_window *mag = new mag_window("magnify",450,300,20,370);
  XWindowAttributes attr;
  XGetWindowAttributes(display,mag->Win,&attr);
  if (attr.map_state == IsViewable) { 
    mag->Unmap();  grab_wi->magw = 0; }
  else { mag->RealizeChildren(); grab_wi->magw = mag; }
}


// ***** class mag_window ********

mag_window::mag_window(char *name, int w, int h, int x, int y) : 
  main_window(name,w,h,1,x,y) {
  mag_im = new mapped_image(w,h,True_Color,True);
  fx = 15; fy = 15;
}

// display magnified region of cap_buf (bw = image width offset)
void mag_window::magnify(unsigned short *cap_buf, window *src_win, 
			 int x, int y) {
  unsigned short *dst_buf = (unsigned short*) mag_im->ximage->data;
  sx = width/fx; sy = height/fy;
  int bw = src_win->width,i,j;
  char wmstr[100]; 
  sprintf(wmstr,"mag (%d,%d) - (%d,%d)",x,y,x+sx-1,y+sy-1);
  XStoreName(display,Win,wmstr); // set new WM name : x,y-range
  for (j=0; j < sy; j++) { // j = line nr in src
    unsigned short tmp[width]; // temp store one line for the dest image
    for (i=0; i < sx; i++) { // i = column nr in src
      unsigned short us = cap_buf[i+x+(y+j)*bw]; // the pixel value to magnify
      for (int ii=0; ii < fx; ii++) tmp[i*fx + ii] = us;
    }
    for (int jj=0; jj < fy; jj++) 
      memcpy(dst_buf+(j*fy+jj)*width,tmp,2*width);
  }
  mag_im->PutImage(Win,0,0,width,height);

  // draw the luminance values into pixels
  for (j=0; j < sy; j++)  for (i=0; i < sx; i++) {
    char val[10];
    unsigned short us = cap_buf[i+x+(y+j)*bw]; 
    unsigned char lum = lum565(us);
    val[0] = '0' + lum / 10; val[1] = '0' + lum % 10; 
    // sprintf(val,"%2d",lum565(us));
    
    XDrawString(display,Win,gc_clear,i*fx+2,j*fy+12,val,2);
  }

}

// ****** class "view_window" ***** 
// size is determined from globals bi...
view_window::view_window(window &parent, int x, int y, info_window *infw) :
  window(parent,biWidth,biHeight,x,y,0), infw(infw) { 
  // info window at bottom
  selection_mask |= ButtonPressMask | ButtonReleaseMask | PointerMotionMask;
  // Cursor grab_cursor = XCreateFontCursor(display,24); // XC_circle
  // XDefineCursor(display,Win,grab_cursor);
  redraw_chain = NULL;
  magw = NULL; // the magnifier window
  create_buffers(biWidth, biHeight);
  mark.x = mark.y = 0;
}

static Bool zieh_mode = False; // draw the mag_window with cursor

// info about cursor point coloring
void view_window::Motion_CB(XMotionEvent ev) {
  short c,b,r,g;  // first part : values from capture buffer  
  // printf("%d %d,",ev.x,ev.y); fflush(stdout);
  c = cap_buf[ev.x + ev.y*biWidth]; 
  b = b565(c);
  g = g565(c);
  r = r565(c);
  short lum = r+g+b; // luminance value

  int rf = (100*r)/31, gf = (100*g)/63, bf = (100*b)/31; // procent values
  sprintf(infw->info,"(%3d,%3d) c=(%2d,%2d,%2d) %3d (%2d,%2d,%2d)",
	  ev.x,ev.y,r,g,b,lum,rf,gf,bf);
  
  infw->redraw();

  // magnifier managing
  static Bool old_drawn = False;
  if (old_drawn) XDrawRectangle(display,Win,gc_rubber,mark.x,mark.y,sx,sy);
 
  if (zieh_mode) { 
    if (magw) { 
      magw->magnify(cap_buf,this,ev.x,ev.y);
      sx = magw->sx; sy = magw->sy;
    } else sx = sy = 20;
    mark.x = ev.x; mark.y = ev.y;
    XDrawRectangle(display,Win,gc_rubber,mark.x,mark.y,sx,sy);
    old_drawn = True;
  } else old_drawn = False;
  view_motion.call(&ev);  // here : call frame_replayer
}

void (*BP_view_cb)(int, int) = 0; // initial define 

// if a magnify window is present toggle zieh_mode, else call "BPcb"
void view_window::BPress_CB(XButtonEvent ev) {
  if (magw) zieh_mode = ! zieh_mode; else
    if (BP_view_cb) (*BP_view_cb)(ev.x,ev.y);// should be replaced by new BP_cb
  view_bpress.call(&ev); // here : call frame_replayer
}

void view_window::redraw() { 
  if (ximage == NULL) return;
  int ww = MIN(width,ximage->width),
      wh = MIN(height,ximage->height);

  act_image->PutImage(Win,0,0,ww,wh);
 
  char * format = (width < 300) ? "%3d %3d %.2f %.2f %d": 
                                    "w = %d h = %d tc = %.2f td = %.2f %d"; 
  int sharp = 0; // = sharpness();
  sprintf(infw->info,format,ww,wh,0.01*tcap,0.01*tdraw,sharp);
  infw->redraw();
  if (redraw_chain) (*redraw_chain)();
}

void view_window::Expose_CB(XExposeEvent ev) {
  XPutImage(display,Win,gc_copy,ximage,ev.x,ev.y,ev.x,ev.y,
	    ev.width,ev.height); 
} 

// create a new object of mapped_image of requested size, 
// delete the old one (if present)
// set the globals "cap_buf" and "ximage", "act_image"
void create_buffers(int w, int h) {
  if (act_image) delete act_image;
  act_image = new mapped_image(w,h, True_Color);
  cap_buf = act_image->buf_16;
  ximage = act_image->ximage;
}  

// adjusting the size of the main window to fit an image ww*hh
void adjust_size(int ww, int hh) {
  window *mw = grab_wi->mainw; // not yet consistent
  mw->resize(ww,hh+40);
}

// specialized main_window creates menubar, view_window, info window
// handles resize events to set resolution (create buffers)
// and handle possibly actions from radio_menu
display_main::display_main(char *Name, int w, int h) : main_window(Name,w,h) {
  menub = new menu_bar(*this,width,20,0,0,30); // top
  infw = new info_window(*this,width,20,0,height-20); // bottom
  view = new view_window(*this,0,20,infw);
}

void display_main::resize(int w, int h) {  
  if ((w != width) || (h != height)) {
    width = w; height = h;     
    biWidth = w & ~1; // make even 
    biHeight = h - 40 ;  // - display-line - menu_bar
    create_buffers(biWidth,biHeight);
    view->resize(biWidth,biHeight); // room for menu_bar
    menub->resize(w,20);     
    infw->resize(w,20); 
    XMoveWindow(display,infw->Win,0,height-20); 
    XResizeWindow(display,Win,width,height); 
  }
}
